import numpy as np
import pandas as pd
from Bio.Cluster import kcluster
from sklearn.cluster import spectral_clustering, SpectralBiclustering, KMeans
from sklearn.metrics import calinski_harabasz_score as ch_score
from sklearn.model_selection import KFold
from .analysis import ClusterResult, affinity_matrix
from .utils import *

__all__ = ['CrossCluster','build_results']

class CrossCluster():
    """Cross clustering of mixed expression matrix.
    if `preset` is not set, then must use :func:`CrossCluster.build_from_func` or :func:`CrossCluster.build_from_model` to set the function to cluster. 
    """    
    def __init__(
        self, preset=None, 
        use_affinity=False, verbose=True,
        **kwargs) -> None:
        """

        :param preset: presetted mod,accept one of below: `kmeans`,`spectrum` and `spectral_bicluster`, defaults to None
        :type preset: str, optional
        :param use_affinity: fit calculate affinity to cluster, defaults to False
        :type use_affinity: bool, optional
        :param verbose: defaults to True
        :type verbose: bool, optional
        """        
        super().__init__()
        self.__name__ ='CrossCluster'
        self.use_aff = use_affinity
        self.verbose = verbose
        self.properties = {}
        self.kwargs = kwargs
        self.presetting = preset

    def build_from_func(self,func):
        """build this class from function

        :param func: functions to cluster
        :type func: function
        """        
        self.cluster_func = func
    
    def build_from_model(self,model,model_attr='label_',**calling_kwargs):
        """build this class from model

        :param model: model to cluster, must have `model_attr` attribute and `fit` method
        :type model: object
        :param model_attr: defaults to 'label_'
        :type model_attr: str, optional
        """        
        def calling(data):
            model.fit(data,**calling_kwargs)
            return getattr(model,model_attr)
        self.cluster_func = calling
    
    def __call__(self,before_cluster_df,**kwargs) -> dict:
        """call this class from built function or model to cluster

        :param before_cluster_df: dataframe to cluster
        :type before_cluster_df: pandas.DataFrame or np.array
        :param kwargs: kwargs to cluster function
        :return: cluster result including `cluster_label` and `cluster_properties`
        :rtype: dict
        """        
        self.kwargs.update(kwargs)
        self.preset()

        if self.use_aff:
            dist_type = kwargs.get("dist_type", "a")
            weight = kwargs.get("weight", [1]*before_cluster_df.shape[1])

            printv("Using affinity matrix will cost more time and meomory.",verbose=self.kwargs.get('verbose',True))
            if dist_type not in ["a", "x", "cosine", "correlation"]:
                print("Affinity is not distance, make sure your distance type is true.(Correlation-like should be used, like 'a','x','cosine','correlation'). Or it will cost much time and get no result.")
            before_cluster_df = affinity_matrix(data=before_cluster_df, dist_type=dist_type, type="affinity", weight=weight)

        labels,properties = self.cluster_func(before_cluster_df,**kwargs)
        self.properties.update(properties)
        return {'cluster_label':labels, 'cluster_properties':self.properties}

    def preset(self):
        """preset the function to cluster

        :raises NotImplementedError: `preset` is not implemented 
        """        
        preset = self.presetting
        if preset is None:
            return 
        if type(preset) is str:
            if preset == 'kmeans':
                self.preset_kmeans()
            elif preset == 'spectrum':
                self.preset_spectrum()
            elif preset == 'spectral_bicluster':
                self.preset_spectral_bicluster()
            else:
                raise NotImplementedError(f'preset = {preset} is not implemented')
        elif type(preset) is function:
            self.build_from_func(preset)
        else:
            try:
                self.build_from_model(preset)
            except Exception as e:
                TypeError(f"preset should be str/function/model;current {type(preset)} and cause {e}")

    def preset_kmeans(self):
        """preset kmeans, default kwargs: `n_clusters`=8,`method`='a', `npass`=20
        """        
        def kcluster_calling(data,**kwargs):
            method = kwargs.get("method", "a")
            dist_type = kwargs.get("dist_type", "a")
            weight = kwargs.get("weight", [1]*data.shape[1])
            n_clusters = kwargs.get("n_clusters", 8)
            npass = kwargs.get("npass", 20)
            label, error, nfound = kcluster(
                data=data, method=method,
                nclusters=n_clusters, weight=weight,
                npass=npass, dist=dist_type)
            return label,{'error':error,'method':f'kmeans_{method}','dist_type':dist_type,'n_clusters':n_clusters,'npass':npass}
        self.build_from_func(kcluster_calling)

    def preset_spectrum(self,):
        """preset spectral clustering, default kwargs: `n_clusters`=8,`use_aff`=True, `n_init`=10
        """        
        def spectrum_calling(data,**kwargs):
            n_clusters = kwargs.get("n_clusters", 8)
            n_component = max(n_clusters) if n_component is None else n_component
            n_init = kwargs.get("n_init", 10)
            weights = [1]*data.shape[1] if weights is None else weights
            return spectral_clustering(
                n_clusters=n_clusters,
                n_component=n_component,
                n_init=n_init,
                affinity=data,
            ),{'n_clusters':n_clusters,'n_component':n_component,'n_init':n_init,'method':'spectrum',}
        self.use_aff=True
        self.build_from_func(spectrum_calling)

    def preset_spectral_bicluster(self):
        """preset spectral bicluster, default kwargs: `n_clusters`=8,`use_aff`=True, `n_init`=10,`method`='log',`n_component`=`m_cluster`, `use_aff`=True
        """        
        def spectral_bicluster_calling(data,**kwargs):
            n_clusters = kwargs.get("n_clusters", 8)
            n_component = max(n_clusters) if n_component is None else n_component
            n_init = kwargs.get("n_init", 10)
            model = SpectralBiclustering(
                n_clusters=n_clusters, method="log", n_components=n_component, n_init=n_init)
            model.fit(data)
            return model.row_labels_,{'method':model,'n_clusters':n_clusters,'n_component':n_component,'n_init':n_init}
        self.use_aff = True
        self.build_from_func(spectral_bicluster_calling)


@kw_decorator(kw="Result")
def build_results(
        cluster_label, mixed_genes, cluster_properties, df_exp, df_pheno,
        before_cluster_df, order_rule="input", verbose=True):
    """building results of :class:`CrossCluster` to :class:`MATTE.analysis.ClusterResult`
    **decorated by :func:`MATTE.utils.kwdecorator`**

    :param cluster_label: label of clustering
    :type cluster_label: array like
    :param mixed_genes: index of mixed genes(in the order of cluster_label)
    :type mixed_genes: array like
    :param cluster_properties: information get from clustering
    :type cluster_properties: dict
    :param df_exp: inputs to all pipeline
    :type df_exp: `pandas.DataFrame`
    :param df_pheno: inputs to all pipeline
    :type df_pheno: `pandas.Series`
    :param before_cluster_df: inputs to :class:`CrossCluster`
    :type before_cluster_df: `pandas.DataFrame`
    :param order_rule: how to order module. 'input'(mean exp) or 'size'(module genes) or function; defaults to "input"
    :type order_rule: str, optional
    :param verbose: defaults to True
    :type verbose: bool, optional
    :return: results
    :rtype: :class:`MATTE.analysis.ClusterResult`
    """    
    printv("building cluster results", verbose=verbose)
    genes = mixed_genes
    assert len(genes) == len(
        cluster_label), "Contains duplicated genes, please check inputs."
    res = pd.Series(index=genes, data=cluster_label)
    cluster_properties["score"] = ch_score(before_cluster_df, res)
    before_cluster_df = pd.DataFrame(before_cluster_df, index=genes)
    res = ClusterResult(
        cluster_res=res, df_exp=df_exp, df_pheno=pd.DataFrame(
            df_pheno).iloc[:, 0],  # Only first column will be sabved in Result.pheno
        before_cluster_df=before_cluster_df,
        cluster_properties=cluster_properties,
        order_rule=order_rule)

    return res



#-- ploting
import matplotlib.pyplot as plt
import matplotlib
from matplotlib import cm
import seaborn as sns
# -- data process
import pandas as pd
import numpy as np
#-- models
from sklearn.metrics import calinski_harabasz_score as ch_score
from sklearn.metrics import davies_bouldin_score as db_score
from sklearn.decomposition import PCA
from scipy import stats
from statsmodels.stats.multitest import fdrcorrection
from scipy.spatial.distance import pdist, squareform
#-- tools
from copy import deepcopy
from tqdm import trange
from .utils import *
__all__=["WeightedDataFrame","ClusterResult","Fig_Fuction","Fig_SampleFeature","FunctionEnrich"]

class WeightedDataFrame(pd.DataFrame):
    """ A weighted dataframe inherit `pandas.DataFrame`, and add a new attribute `weight` to store the weight of each column.
    """    
    def __init__(self, weight=None, data=None, index=None, columns=None, dtype=None, copy=None) -> None:
        """except weight, others parameters are same to pandas.DataFrame.

        :param weight: weight of dataframe, defaults to None
        :type weight: array-like, optional
        """        
        super().__init__(data=data, index=index, columns=columns, dtype=dtype, copy=copy)
        self.weight = pd.Series(weight, index=columns)
        if len(self.weight) == 0:
            self.weight = [1]*self.values.shape[1]

    def weight_distance(self, metric='c', **kwargs):
        """Calculate weighted distance.

        :param metric: The distance metric to use., defaults to 'c'
        :type metric: str, optional

        .. note:: calling from different module can be very different for their definition is not same.
        The distance function can be :
        --- calling from scipy.spatial.pdist ---
        'braycurtis', 'canberra', 'chebyshev',
        'cityblock', 'correlation', 'cosine', 'dice', 'euclidean', 'hamming',
        'jaccard', 'jensenshannon', 'kulsinski', 'mahalanobis', 'matching',
        'minkowski', 'rogerstanimoto', 'russellrao', 'seuclidean', 'sokalmichener',
        'sokalsneath', 'sqeuclidean', 'yule'.
        --- calling from Bio.cluster.distancematrix ---
        'e': Euclidean ;
        'b': City-block .
        'c': Pearson correlation;
        'a': Absolute of Pearson correlation;
        'u': Uncentered Pearson correlation;
        'x': Absolute of Uncentered Pearson correlation;
        's': Spearman's correlation;
        'k': Kendall's τ.
        
        :return: distance matrix
        :rtype: numpy.array
        """
        if metric not in ["a", "x"]:
            print(
                "Calculate distance will not return negtive value. Result will be absolute!")
        if len(metric) > 1:
            return squareform(
                pdist(self.values, metric=metric,
                      w=self.weight.values, **kwargs)
            )
        else:
            return abs(affinity_matrix(data=self.values,
                                             dist_type=metric, weight=self.weight.values))


class ClusterResult():
    """A class to store the result of clustering. contains inputs and results.
    """    
    def __init__(self, cluster_res, before_cluster_df,df_exp,df_pheno,cluster_properties,order_rule="input") -> None:
        """

        :param cluster_res: cluster result of :class:`MATTE.cluster.CrossCluster`
        :type cluster_res: array-like
        :param before_cluster_df: inputs to :class:`MATTE.cluster.CrossCluster`
        :type before_cluster_df: numpy.array or pandas.DataFrame
        :param df_exp: inputs to all the pipeline
        :type df_exp: pandas.DataFrame
        :param df_pheno: inputs in all the pipeline
        :type df_pheno: pandas.Series
        :param cluster_properties: cluster result of :class:`MATTE.cluster.CrossCluster`
        :type cluster_properties: dict
        :param order_rule: order modules by the rule, defaults to "input"
        :type order_rule: str, optional
        """        
        self.n_cluster = cluster_res.max()+1
        self.cluster_properties = cluster_properties  # prperty record the score and loss of the cluster results
        self.label = pd.Series([i.split("@")[1]
                                for i in cluster_res.index]).unique()
        self.df_exp = df_exp
        self.pheno = df_pheno
        self._reorder(before_cluster_df, cluster_res,by=order_rule)
        self._fixingRes(cluster_res, label=self.label)
        self._JMatrix()


    # --- built-in methods

    def _reorder(self, before_cluster_df, cluster_res, by="input"):
        """reorder cluster by data.

        :param before_cluster_df: preprcossed exp df.
        :type before_cluster_df: pd.DataFrame
        :param cluster_res: cluster result of :class:`MATTE.cluster.CrossCluster`
        :type cluster_res: array-like
        :param by: order rule, defaults to "input". 'size' is also valid option. 
        :type by: str,or function optional
        """
        if by=="input":
            assert len(cluster_res) == before_cluster_df.shape[0]
            df_input_ = deepcopy(before_cluster_df)
            df_input_["group"] = cluster_res
            rank = pd.Series(df_input_.groupby(
                "group").median().median(axis=1).rank(), dtype=int)
        elif by=="size":
            rank = pd.Series(cluster_res.value_counts().rank(ascending=False),dtype=int)
        elif type(by) == type(np.sum):
            rank = pd.Series(by(cluster_res,before_cluster_df),dtype=int)
        for i in range(cluster_res.shape[0]):
            cluster_res[i] = rank[cluster_res[i]]-1


    def _fixingRes(self, cluster_res, label, sep="@"):
        """from cluster results to build a dataframe, remove special sperate @.

        :param cluster_res: cluster result of :class:`MATTE.cluster.CrossCluster`
        :type cluster_res: array-like
        :param label: label of cluster
        :type label: array-like
        :param sep: defaults to "@"
        :type sep: str, optional
        """        
        # fix result. and sperate genes.
        genes_mixed = cluster_res.index
        genes = pd.Series([i.split(sep)[0] for i in genes_mixed]).unique()
        fixed_res = pd.DataFrame(index=genes, columns=label)
        for i in range(cluster_res.shape[0]):
            group = cluster_res[i]
            gene = genes_mixed[i]
            gene, spe = gene.split(sep)
            if spe in label:
                fixed_res.loc[gene, spe] = group
            else:
                raise ValueError(
                    "Label is not consistent with input.(or sep is wrong)")

        # genes that not in same group.
        matched = [(fixed_res.iloc[i, :] == fixed_res.iloc[i, 0]).all()
                   for i in range(fixed_res.shape[0])]
        fixed_res["matched"] = matched

        self.res = fixed_res

    def _JMatrix(self):
        """calculate J matrix(cross tabulation), showing the relationship between clusters of phenotypes.
        """        
        JM = np.zeros(shape=[self.n_cluster]*len(self.label))
        try:
            for i in range(self.res.shape[0]):
                JM[tuple(self.res.iloc[i, :-1])] += 1
        except Exception as e:
            print(e)
            print("Using index:", tuple(self.res.iloc[i, :-1]))
            print("Using result", self.res.head())
            #print("Using matrix",JM,JM.shape)
            raise
        self.JM = JM

    def _module_genes(self):
        """get genes in each module.And store in attibute :attr:`ClusterResult.module_genes`

        :return: keys : module id | values : gene list
        :rtype: dict
        """        
        module_genes = {}
        for ind in zip(*np.where(self.JM > 0)):
            genes = self.res.index[(
                self.res.iloc[:, 0:-1] == list(ind)).all(axis=1)]
            if (pd.Series(ind) == ind[0]).all():
                continue
            module_genes["M"+".".join(list(map(str, list(ind))))] = genes
        self.module_genes = module_genes
        return module_genes

    def summary(self, fig=True):
        """summary of cluster results,print :attr:`ClusterResult.cluster_properties`, and plot J matrix.

        :param fig: whether print figures or not, defaults to True
        :type fig: bool, optional
        :return: summary of cluster results
        :rtype: figures or None
        """        
        print("# --- Number of genes:")
        print("Same Module Genes:", self.res["matched"].sum())
        print("Different Module Genes:",
              self.res.shape[0] - self.res["matched"].sum())
        print("# --- clustering score:")
        for k, v in self.cluster_properties.items():
            print(k, v)
        if fig:
            f1 = self.Vis_Jmat()
            print("# --- samples' distribution:")
            sf = self.SampleFeature(df_exp=self.df_exp)
            f2 = Fig_SampleFeature(
                sf, labels=self.pheno,
                model=PCA(),metric="euclidean",
                dpi=150)
            return f1, f2

    def Vis_Jmat(self):
        """plot J matrix.(heatmap)

        :return: figure
        :rtype: matplotlib.figure.Figure
        """        
        JM = self.JM
        f1 = plt.figure(dpi=300)
        sns.heatmap(
                data=JM, fmt=".0f",
                cmap="RdPu", square=True,
                annot=True, annot_kws={"fontsize": 8})
        plt.title("Genes' distribute")
        plt.xlabel(self.label[-1].split("|")[-1])
        plt.ylabel(self.label[-2].split("|")[-1])
        return f1

    # Cluster Score Calculation
    def MCScore(self, method=ch_score):
        """calculation subscore:the loss between each mc. the result stored in `self.property["MCScore"]

        :param method: method used to calculate, annother recommend function:`sklearn.metrics.davies_bouldin_score`, defaults to `sklearn.metrics.calinski_harabasz_score`.
        :type method: _type_, optional
        """        
        self.property["MCScore"] = method(self.df_exp, [str(
            self.res.iloc[i, 0:2].values) for i in range(self.res.shape[0])])

    def MCFeature(self, df_exp=None, module_genes=None, model=None):
        """calculation group features.using PCA for it reserving more information than others.

        .. note:: 
            if n_components >=2, MCs' eigenvector can reserve more than one to make explained ratio >=80%.

        :param df_exp: defaults to None, :attr:`ClusterResult.df_exp`
        :type df_exp: pd.DataFrame, optional
        :param module_genes: the mapping of module and genes. defaults to None, result of function :func:`MATTE.analysis.ClusterResult._module_genes`, 
        :type module_genes: dict, optional
        :param model: the model to fit transform data.needs attribute `fit_transform`, defaults to None(`sklearn.decomposition.PCA(n_components=1)`)
        :type model: object, optional
        :return: group_feature: `dict`,keys : module id | values : eigenvector, group_weight: `dict`,keys : module id | values : lambda of model 
        :rtype: tuple
        """        
        if df_exp is None:
            df_exp = self.df_exp
        if module_genes is None:
            module_genes = self._module_genes()

        group_feature = {}
        group_weight = {}
        for mc, genes in module_genes.items():
            pca = PCA(n_components=1) if model is None else model
            feature = pca.fit_transform(
                df_exp.loc[genes, :].T)
            all_explained = 0
            for ind, ratio in enumerate(pca.explained_variance_ratio_):
                all_explained += ratio
                if all_explained >= 0.8:
                    break

            group_feature[mc] = feature[:, 0:ind+1]
            group_weight[mc] = pca.explained_variance_ratio_[:ind+1]

        return group_feature, group_weight

    # Samples' feature and phenotype's signature.

    def SampleFeature(self, df_exp=None, module_genes=None, return_df=True, **kwargs):
        """Calculate Samples' module feature.(average default)

        :param df_exp: defaults to None, :attr:`ClusterResult.df_exp`
        :type df_exp: pd.DataFrame, optional
        :param module_genes: the mapping of module and genes. defaults to None, result of function :func:`MATTE.analysis.ClusterResult._module_genes`, 
        :type module_genes: dict, optional
        :param return_df: decide return. If False, then will store result in :attr:`ClusterResult.samples_feature`, defaults to True
        :type return_df: bool, optional
        :return: depended by the arg:return_df
        :rtype: samples_feature is a `WeightedDataFrame` object, whose index is samples id and columns is module id. Module Weights are calculated by function `MCFeature`.
        """        
        if df_exp is None:
            df_exp = self.df_exp

        def lst_sum(lst):
            ret = []
            for l in lst:
                ret.extend(l)
            return ret

        module_genes = self._module_genes() if module_genes is None else module_genes
        group_feature, group_weight = self.MCFeature(
            df_exp=df_exp, module_genes=module_genes, **kwargs)
        samples_feature = WeightedDataFrame(
            index=df_exp.columns, data=np.hstack(
                tuple(group_feature.values())),
            columns=lst_sum(
                [
                    [i+"_"+str(_)] for i in group_feature.keys()
                    for _ in range(group_feature[i].shape[1])]),
            weight=np.hstack(list(group_weight.values()))
        )

        if return_df:
            return samples_feature
        else:
            self.samples_feature = samples_feature

    def PhenoMCCorr(self, sample_feature, pheno_series, pheno=None):
        """Calculate the correlation between Phenotype and MC's Eigenvector and test the correlation and return p-value

        :param sample_feature: can be calculated by the function :func:`ClusterResult.SampleFeature`
        :type sample_feature: :class:`MATTE.WeightedDataFrame`
        :param pheno_series: specify samples' phenotype
        :type pheno_series: pd.Series
        :param pheno: choose which pheno the calculate, defaults to None (the first one of `pheno_series`)
        :type pheno: any, optional
        :return: `MC_corr`: a `pd.Series` whose index is Module ID, and value is the pearson correlation. `corr_p`: p_value of correlation(Student t test)
        :rtype: tuple
        """        
        def covPvalueStudent(cor, n_samples):
            T = (n_samples - 2)**0.5 * cor/((1-cor**2)**0.5)
            return 2*stats.t.sf(abs(T), n_samples-2)

        if pheno is None:
            pheno = pheno_series.unique()[0]
        MC_corr = pd.Series(index=sample_feature.columns, dtype=float)
        for i in range(sample_feature.shape[1]):
            MC_corr[i] = np.corrcoef(
                sample_feature.iloc[:, i],
                pheno_series[sample_feature.index] == pheno)[0, 1]
        corr_p = pd.Series(index=MC_corr.index, data=covPvalueStudent(
            MC_corr, n_samples=len(MC_corr)))
        return MC_corr, corr_p

    def MCkwtest(self, sample_feature, pheno_series):
        """Making the Kruskal-Wallis test on each MC.
        (if Each module's Eigenvector is different)

        :param sample_feature: can be calculated by the function :func:`MATTE.analysis.ClusterResult.SampleFeature`
        :type sample_feature: :class:`MATTE.WeightedDataFrame`
        :param pheno_series: specify samples' phenotype
        :type pheno_series: pd.Series
        :return: `kws`: a `pd.Series` whose index is Module ID, and value is the p-value.
        :rtype: pd.Series
        """        
        Sig = [
            sample_feature.loc[pheno_series[pheno_series == li].index, :]
            for li in pheno_series.unique()
        ]

        kws = pd.Series(index=sample_feature.columns, data=0.0)
        for i in range(len(kws)):
            kw = stats.kruskal(*[sig.iloc[:, i] for sig in Sig]).pvalue
            kws[i] = kw
        return kws

    def ModuleSNR(self, sample_feature, pheno=None):
        """Calculate the SNR of each module.

        :param sample_feature: can be calculated by the function :func:`MATTE.analysis.ClusterResult.SampleFeature`
        :type sample_feature: :class:`MATTE.WeightedDataFrame`
        :param pheno: the label of samples, defaults to None (set to be :attr:`ClusterResult.pheno`)
        :type pheno: _type_, optional
        :return: SNR, a `pd.Series` whose index is Module ID, and value is the SNR.
        :rtype: pd.Series
        """        
        if pheno is None:
            pheno = self.pheno
        SNR = pd.Series(index=sample_feature.columns, dtype=float)
        assert pheno.unique().size == 2, "The pheno should be binary"
        pos,neg = pheno.unique()
        pos_samples = pheno[pheno==pos].index
        neg_samples = pheno[pheno==neg].index
        for i in sample_feature.columns:
            mean_pos = np.mean(sample_feature.loc[pos_samples, i])
            mean_neg = np.mean(sample_feature.loc[neg_samples, i])
            sd_pos = np.std(sample_feature.loc[pos_samples, i])
            sd_neg = np.std(sample_feature.loc[neg_samples, i])
            SNR[i] = np.abs(mean_pos-mean_neg)/ (sd_neg+sd_pos+1e-10)
        return SNR.sort_values(ascending=False)

# Showing

def Fig_SampleFeature(
        sample_feature, labels, color=None,
        model=None, weighted_distcance=False, metric='euclidean', **fig_args):
    """showing the fig of samples reprented by MCs.

    :param sample_feature: index : samples' id | columns : modules' id . the result of :func:`ClusterResult.SampleFeature` or just using some genes to explain a sample.
    :type sample_feature: `WeightedDataFrame` or `pd.DataFrame` 
    :param labels: index : samples' id | columns : phenotype
    :type labels: pd.Series
    :param color: the second label used, and in the form of color. the first one will be in the form of shape., defaults to None
    :type color: array like, optional
    :param model: the model used to present samples in the two dimensions., defaults to None (`sklearn.manifold.PCA(n_component=2)`)
    :type model: object, optional
    :param weighted_distcance: whether to used weighted distance. if sample_feature is `pd.DataFrame`, it should be set as False., defaults to False
    :type weighted_distcance: bool, optional
    :param metric: used to calculate the distance., defaults to 'euclidean'
    :type metric: str, optional
    :return: the figure
    :rtype: `plt.figure`
    """
    model = PCA(n_components=2) if model is None else model
    if weighted_distcance:
        psf = abs(sample_feature.weight_distance(metric=metric))
        psf = model.fit_transform(psf)
    else:
        psf = model.fit_transform(sample_feature)

    if not hasattr(model, "fit_transform"):
        raise TypeError("model has no method named fit_transform.")

    Samples = sample_feature.index
    labels = labels[Samples]
    if color is None:
        fig = plt.figure(**fig_args)
        for i, l in enumerate(labels.unique()):
            plt.scatter(
                x=psf[labels == l][:, 0], y=psf[labels == l][:, 1],
                alpha=.8, label=l, color=matplotlib.cm.Paired(i/len(labels.unique())))
        plt.legend(loc='best', shadow=False, scatterpoints=1)
    else:
        color = color[Samples]
        fig = plt.figure(**fig_args)
        markers = ["o", "x", ",", "v", "^", "<", ">", "*", "+", ".",
                   "1", "2", "3", "4", "8", "s", "p", "h", "d", "D", "|", "_"]
        for i, l in enumerate(labels.unique()):
            plt.scatter(
                x=psf[labels == l][:, 0], y=psf[labels == l][:, 1],
                alpha=.8, label=l, c=color[labels == l], marker=markers[i])
        plt.legend(loc='best', shadow=False, scatterpoints=1)
        plt.colorbar()

    return fig


def FunctionEnrich(annote_file, gene_set, category_seperate_cal=True):
    """perform GO enrichment

    :param annote_file: Function annotation file dir or dataframe
    .. note:: tab-seperate file with columns:["Term_ID","GeneID","Term","Category"];File can be downloaded from https://ftp.ncbi.nih.gov/gene/DATA/
    :type annote_file: `str` or `pd.DataFrame`
    :param gene_set: gene set to analysis
    :type gene_set: array-like
    :param category_seperate_cal: wheather to calculate fdr in each category or mixed., defaults to True
    :type category_seperate_cal: bool, optional

    :return: items,catogory,enriched items number,backgroud item number,p value,fdr
    :rtype: pd.DataFrame
    """    
    if type(annote_file) == str:
        annot = pd.read_table(annote_file, index_col="Term_ID")
    elif type(annote_file) == pd.DataFrame:
        annot = deepcopy(annote_file)
    assert set(["GeneID", "Term", "Category"]) <= set(annote_file.columns)
    annot.index = annote_file["Term_ID"]
    annot["GeneID"] = pd.Series(annot["GeneID"], dtype=type(gene_set[0]))
    gene_set = set(gene_set) & set(annote_file["GeneID"])
    gene_set = pd.Series(list(gene_set), dtype=object)
    try:
        assert len(gene_set) != 0
    except:
        raise ValueError("No genes in the annotation file.")

    all_items = annot[["Term", "Category"]].drop_duplicates()
    enriched = pd.Series(index=all_items.index, data=0, dtype=int)
    backgroud = pd.Series(index=all_items.index, data=0, dtype=int)
    p_value = pd.Series(index=all_items.index, data=0.0, dtype=float)
    n_background = annot["GeneID"].unique().size
    n_set = len(gene_set)
    term_genes = {}
    for i in trange(all_items.shape[0]):
        term_id = all_items.index[i]
        all_gene_item_related = pd.Series(annot.loc[term_id, "GeneID"])
        gene_list = all_gene_item_related.unique()
        gene_list_target = pd.Series(
            list(set(gene_list) & set(gene_set)), dtype=object)

        B = gene_list.size
        b = gene_list_target.size

        term_genes[term_id] = gene_list_target
        enriched[term_id] = b
        backgroud[term_id] = B

        if b != 0:
            p_value[term_id] = stats.hypergeom.sf(b-1, n_background, B, n_set)
        else:
            p_value[term_id] = 1.0

    all_items["n_enriched"] = enriched
    all_items["n_backgroud"] = backgroud
    all_items["p_value"] = p_value
    all_items["fdr"] = pd.NA

    if category_seperate_cal:
        for cat in all_items["Category"].unique():
            idx = all_items[all_items["Category"] == cat].index
            all_items.loc[idx, "fdr"] = fdrcorrection(
                all_items.loc[idx, "p_value"].values)[1]
    else:
        all_items.loc[:, "fdr"] = fdrcorrection(
            all_items.loc[:, "p_value"].values)[1]

    all_items = all_items.sort_values(by=["fdr", "p_value", "n_enriched"])
    all_items['gene_ratio'] = all_items['n_enriched'] / len(gene_set)
    return all_items, term_genes


def Fig_Fuction(df_enrich, color_columns, cmap=cm.Set1, width_height_ratio=2, **figargs):
    """figuring the function enrichment reuslt.
    .. note:: filtering the result first!

    :param df_enrich: the result of :func:`FunctionEnrich`
    :type df_enrich: `pd.DataFrame`
    :param color_columns: used to shown in color, 'fdr' or 'p_value'
    :type color_columns: `str`
    :param cmap: chosen from `matplotlib.cm`, defaults to cm.Set1
    :type cmap: object, optional
    :param width_height_ratio: defaults to 2
    :type width_height_ratio: int, optional
    :return: the figure
    :rtype: plt.figure
    """
    lable_max = int(np.log2(df_enrich[color_columns].min()))
    fig = plt.figure(constrained_layout=True, **figargs)
    gs = fig.add_gridspec(
        df_enrich.shape[0], df_enrich.shape[0]*width_height_ratio)
    used_gride = 0
    axes = []
    for n, cat in enumerate(df_enrich["Category"].unique()):
        ctarget = df_enrich[df_enrich["Category"] == cat]
        ax = fig.add_subplot(gs[used_gride:used_gride+ctarget.shape[0], :])
        used_gride += ctarget.shape[0]
        ax.barh(
            width=ctarget["gene_ratio"],
            y=range(ctarget.shape[0]),
            color=[cm.coolwarm(np.log2(i)/lable_max) for i in ctarget[color_columns]])
        ax.set_yticks(range(ctarget.shape[0]))
        ax.set_yticklabels(ctarget.index.values.tolist())
        ax.tick_params("y", labelcolor=cmap(n))
        ax.text(
            1.02, 0, cat,
            transform=ax.transAxes, rotation=90, color=cmap(n), verticalalignment='bottom',
            weight='semibold')
        axes.append(ax)

    ## setting all axes in the figure share the x axis
    xlim_max = int(df_enrich['gene_ratio'].max()*10+1)/10
    for ax in axes:
        ax.set_xlim(0, xlim_max)
    
    for ax in axes[:-1]:
        ax.set_xticklabels([])
        ax.set_xticks([])
    plt.xlabel("Gene Ratio")

    cbar_ax = fig.add_axes([1.01, 0.12, 0.03, 0.87])
    cbar = fig.colorbar(cm.ScalarMappable(cmap=cm.coolwarm), cax=cbar_ax)
    cbar.set_ticks(ticks=[0, 0.25, 0.5, 0.75, 1])
    cbar.set_ticklabels([0, -lable_max/4, -lable_max /
                         2, -3*lable_max/4, -lable_max])
    cbar.set_label("-log(%s)" % color_columns)

    return fig



import time
import numpy as np
from Bio.Cluster import distancematrix
from functools import wraps
import pandas as pd
from inspect import getfullargspec
__all__ = ["printv","affinity_matrix","kw_decorator"]

def printv(*text,show_time=True,verbose=True):
    """Prints text with time and verbose.

    :param show_time: defaults to True
    :type show_time: bool, optional
    :param verbose: defaults to True
    :type verbose: bool, optional
    """    
    if verbose:
        if show_time: print(time.ctime()+"\t",*text)
        else:print(*text)

def affinity_matrix(data, dist_type,type="distance", **kwargs):
    """Calculate affinity matrix; both implement from :mod:`Bio.Cluster` and :mod:`scipy.spatial.distance` are supported.

    :param data: data to calculate affinity matrix
    :type data: numpy.array or pandas.DataFrame
    :param dist_type: string of distance type
    :type dist_type: str
    :param type: `distance` or `affinity`, defaults to "distance"
    :type type: str, optional
    :return: matrix
    :rtype: np.array
    """    
    aff_matrix = np.zeros(shape=(data.shape[0], data.shape[0]))
    for ind, value in enumerate(
            distancematrix(data=data, dist=dist_type, **kwargs)):
        if len(value) == 0:
            continue
        aff_matrix[ind, 0:len(value)] = value
    aff_matrix += aff_matrix.T
    if type == "distance":
        return aff_matrix
    elif type == "affinity":
        return 1-aff_matrix

def kw_decorator(kw=None):
    """Import decarated function used in pipeline.This allow function to accept more than keyworks arguments it self accepted. and will make function return dict as kw set.

    :param kw: keywords make of function's return dict's keys, defaults to None
    :type kw: str or list, optional
    """    
    def funcdec(func):

        @wraps(func)
        def new_fun(*args,**in_use_args):
            func_args_names = getfullargspec(func).args
            in_use_args = {k:w for k,w in in_use_args.items() if k in func_args_names}
            if kw is None:
                return func(*args,**in_use_args)
            elif type(kw) in [list,tuple,np.ndarray,pd.Series]:
                return dict(zip(kw,func(*args,**in_use_args)))
            else:
                return {kw:func(*args,**in_use_args)}
        
        return new_fun

    return funcdec




from .utils import printv, kw_decorator
import pandas as pd
import numpy as np
from sklearn.preprocessing import KernelCenterer
from sklearn.feature_selection import f_classif
from joblib.parallel import Parallel, delayed
from sklearn.preprocessing import normalize,StandardScaler


__all__ = ["RPKM2TPM", "log2transform", "exp_filter", "Kernel_transform", "inputs_check",'normalization']

## --- Preprocess --- ##

@kw_decorator('Checking_status')
def inputs_check(df_exp, df_pheno):
    """Checking inputs if they are correct. decorated by :func:`MATTE.utils.kw_decorator`

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :raises ValueError: If the input dataframe is not correct.
    :return: checking status
    :rtype: dict {`Checking_status`:'OK'}
    """    
    genes = df_exp.index
    samples = df_exp.columns
    phenoes = df_pheno.unique()

    # check gene names and pheno names should not contains '@'
    for i in genes:
        if '@' in i:
            raise ValueError(f"Gene name {i} contains '@'")
    for i in phenoes:
        if '@' in str(i):
            raise ValueError(f"Pheno name {i} contains '@'")
    
    # check samples
    if len(samples) != len(set(samples)):
        raise ValueError("Samples should be unique.")
    if set(list(df_pheno.index)) != set(samples):
        raise ValueError("Samples in df_pheno and df_exp are not consistent.")

    # check genes
    if len(genes) != len(set(genes)):
        raise ValueError("Genes should be unique.")

    assert (df_exp >=0).all().all(), "Expression contains negative values."
    return 'OK'

@kw_decorator(kw='df_exp')
def RPKM2TPM(df_exp):
    """Convert RPKM to TPM. decorated by :func:`MATTE.utils.kw_decorator`

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :return: TPM dataframe.
    :rtype: dict
    """
    df = df_exp.values
    oneM = 1e6
    df_sum = df.sum(axis=0)
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            df[i, j] = (df[i, j]/df_sum[j]) * oneM
    return pd.DataFrame(data=df, columns=df_exp.columns, index=df_exp.index)


@kw_decorator(kw="df_exp")
def log2transform(df_exp):
    """Log2 transform the expression dataframe. decorated by :func:`MATTE.utils.kw_decorator`

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :return: Log2 transformed dataframe.
    :rtype: dict
    """    
    if type(df_exp) == pd.DataFrame:
        return pd.DataFrame(
            data=np.log2(df_exp.values+1),
            index=df_exp.index,
            columns=df_exp.columns
        )
    else:
        return np.log2(df_exp+1)

@kw_decorator(kw="df_exp")
def normalization(df_exp,norm='l1'):
    """Normalize the expression dataframe. decorated by :func:`MATTE.utils.kw_decorator`

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param norm: normalization type,should be one of `l1`,`l2` and `standard`, defaults to 'l1'
    :type norm: str, optional
    :return: Normalized dataframe.
    :rtype: dict
    """    
    cols = df_exp.columns
    rows = df_exp.index      
    if norm in ['l1','l2']:
        return pd.DataFrame(normalize(df_exp,norm=norm),index=rows,columns=cols)
    elif norm == 'standard':
        SD = StandardScaler()
        return pd.DataFrame(SD.fit_transform(df_exp),index=rows,columns=cols)


@kw_decorator(kw='df_exp')
def exp_filter(df_exp,df_pheno,gene_filter=None,filter_args:dict={}):
    """Filter the genes by rule. decorated by :func:`MATTE.utils.kw_decorator`

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :param gene_filter: gene filter rule. `None` will delete some genes with extreme low expression. `f` will filter genes by Anova f value(p=0.05 by default). `function` is also allowed.defaults to None
    :type gene_filter: str or function, optional
    :param filter_args: other args sent to gene filter function, defaults to {}
    :type filter_args: dict, optional
    :return:  filtered dataframe.
    :rtype: dict
    """    
    if gene_filter is None:
        df_exp = df_exp[(df_exp >= 1).any(axis=1)]
    elif gene_filter == 'f':
        thres = filter_args.get('thres',0.05)
        f,p = f_classif(df_exp.T.values,df_pheno.values)
        df_exp = df_exp.iloc[np.where(p<thres)[0],:]
    else:
        df_exp = df_exp[gene_filter(df_exp,df_pheno,**filter_args)]
    return df_exp

## --- Variable Generate  --- ##

def generate_df_exp_mixed(df_exp, df_pheno):
    """Generate expression dataframe with mixed phenotypes. That's the first step of cross clustering. In this step, genes from differnet phenotypes are treated as different genes.

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :return: Expression dataframe with mixed phenotypes, whose index is mixed genes(using `'@'`as a seperate to label genes from phenotypes).
    :rtype: pandas.DataFrame
    """    
    phenoes = df_pheno.unique()
    n_select_samples = df_pheno.value_counts().min()
    df_set = []
    for p in phenoes:
        df_sub = df_exp.loc[:, df_pheno[df_pheno ==
                                        p].index].iloc[:, 0:n_select_samples]
        df_sub.index = [f'{i}@{p}' for i in df_sub.index]
        df_sub.columns = range(n_select_samples)
        df_set.append(df_sub)
    return pd.concat(df_set)


## --- Kernel Transformation --- ##

def outer_subtract(x, absolute=False):
    """Outer subtract the mean of each gene.

    :param x: Expression data
    :type x: numpy.array
    :param absolute: whether after substract to take absolute, defaults to False
    :type absolute: bool, optional
    :return: cross substract mean of each gene.
    :rtype: numpy.array
    """    
    if absolute:
        return np.abs(x[:, np.newaxis] - x[np.newaxis, :])
    else:
        return x[:, np.newaxis] - x[np.newaxis, :]


def variant_rbf_kernel(x, gamma: float = 1.0):
    """Variant RBF kernel.

    :param x: Expression data
    :type x: numpy.array
    :param gamma: rbf kernel parameters, defaults to 1.0
    :type gamma: float, optional
    :return: RBF kernel matrix.
    :rtype: numpy.array
    """    
    assert x.shape[0] == x.shape[1]
    return np.exp(-gamma * x ** 2)


def mean_Kernel_transform(df_exp, df_pheno, genes_mixed, absolute=True,):
    """Mean Kernel transform.

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :param genes_mixed: mixed genes.
    :type genes_mixed: array like
    :param absolute: whether to take absolute, defaults to True
    :type absolute: bool, optional
    :return: Mean Kernel transform matrix.
    :rtype: numpy.array
    """    
    mean_by_pheno = df_exp.groupby(df_pheno, axis=1).mean()
    mixed_genes_mean = mean_by_pheno.stack().reset_index()
    mixed_genes_mean.index = [
        f"{mixed_genes_mean.iloc[i,0]}@{mixed_genes_mean.iloc[i,1]}" for i in range(mixed_genes_mean.shape[0])]
    mixed_genes_mean = mixed_genes_mean.iloc[:, -1].loc[genes_mixed]
    return outer_subtract(mixed_genes_mean.values, absolute)


def meanrbf_Kernel_transform(df_exp, df_pheno, genes_mixed, gamma, absolute=True,):
    """Mean RBF Kernel transform. Is a variant of rbf kernel. 

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :param genes_mixed: mixed genes.
    :type genes_mixed: array like
    :param gamma: rbf kernel parameters, defaults to 1.0
    :type gamma: float, optional
    :param absolute: whether to take absolute in outer substract, defaults to True
    :type absolute: bool, optional
    :return: Mean RBF Kernel transform matrix.
    :rtype: numpy.array
    """    
    mean_by_pheno = df_exp.groupby(df_pheno, axis=1).mean()
    mixed_genes_mean = mean_by_pheno.stack().reset_index()
    mixed_genes_mean.index = [
        f"{mixed_genes_mean.iloc[i,0]}@{mixed_genes_mean.iloc[i,1]}" for i in range(mixed_genes_mean.shape[0])]
    mixed_genes_mean = mixed_genes_mean.iloc[:, -1].loc[genes_mixed]
    return variant_rbf_kernel(outer_subtract(mixed_genes_mean.values, absolute),gamma=gamma)

def localrbf_Kernel_transform(df_exp, df_pheno, genes_mixed,K=7, absolute=True):
    """Local RBF Kernel transform. each genes' gamma is set to be its' distance to K-th nearest neighbor. 

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :param genes_mixed: mixed genes.
    :type genes_mixed: array like
    :param K: nearest neighbor, defaults to 7
    :type K: int, optional
    :param absolute: whether to take absolute or not, defaults to True
    :type absolute: bool, optional
    :return: Mean RBF Kernel transform matrix.
    :rtype: numpy.array
    """    
    mean_by_pheno = df_exp.groupby(df_pheno, axis=1).mean()
    mixed_genes_mean = mean_by_pheno.stack().reset_index()
    mixed_genes_mean.index = [
        f"{mixed_genes_mean.iloc[i,0]}@{mixed_genes_mean.iloc[i,1]}" for i in range(mixed_genes_mean.shape[0])]
    mixed_genes_mean = mixed_genes_mean.iloc[:, -1].loc[genes_mixed]
    mean_mat:np.array = outer_subtract(mixed_genes_mean.values,absolute)

    ## calculation K-th nearest neighbor's distance
    sigmas = pd.Series(index = genes_mixed,dtype=float)
    for i,g in enumerate(genes_mixed):
        ind = mean_mat[i].argsort()[K]
        sigmas[g] = mean_mat[i,ind]
    
    for i in range(mean_mat.shape[0]):
        for j in range(i+1,mean_mat.shape[1]):
            gamma = -1/(sigmas[genes_mixed[i]] * sigmas[genes_mixed[j]])
            mean_mat[i,j] = np.exp(gamma * mean_mat[i,j]**2)
            mean_mat[j,i] = mean_mat[i,j]
    return mean_mat

# Used for any function between two vector
def cdist_generic(dataset1, dataset2, dist_fun, n_jobs=-1, verbose=0,
                  compute_diagonal=True, *args, **kwargs):
    """Compute the distance matrix from two datasets.

    :param dataset1: First dataset.
    :type dataset1: numpy.array
    :param dataset2: Second dataset.
    :type dataset2: numpy.array
    :param dist_fun: Distance function.
    :type dist_fun: function
    :param n_jobs: number of cpus used, defaults to -1
    :type n_jobs: int, optional
    :param verbose: defaults to 0
    :type verbose: int, optional
    :param compute_diagonal: defaults to True
    :type compute_diagonal: bool, optional
    :return: using dist_fun to compute the distance matrix.
    :rtype: numpy.array
    """
    if dataset2 is None:
        matrix = np.zeros((len(dataset1), len(dataset1)))
        indices = np.triu_indices(len(dataset1),
                                  k=0 if compute_diagonal else 1,
                                  m=len(dataset1))
        matrix[indices] = Parallel(n_jobs=n_jobs,
                                   verbose=verbose)(
            delayed(dist_fun)(
                dataset1[i], dataset1[j],
                *args, **kwargs
            )
            for i in range(len(dataset1))
            for j in range(i if compute_diagonal else i + 1,
                           len(dataset1))
        )
        indices = np.tril_indices(len(dataset1), k=-1, m=len(dataset1))
        matrix[indices] = matrix.T[indices]
        return matrix
    else:
        matrix = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(dist_fun)(
                dataset1[i], dataset2[j],
                *args, **kwargs
            )
            for i in range(len(dataset1)) for j in range(len(dataset2))
        )
        return np.array(matrix).reshape((len(dataset1), -1))


## --- Kernel Matrix Process --- ##

def Double_Centering(x: np.array):
    """Double Centering to Kernel Matrix

    :param x: Kernel matrix.
    :type x: np.array
    :return: double centered matrix
    :rtype: numpy.array
    """    
    return x - np.median(x,axis=0) - np.median(x,axis=1) + np.median(x)

## --- Main of Kernel Transformation --- ##

@kw_decorator(kw=['before_cluster_df', 'mixed_genes'])
def Kernel_transform(
    df_exp, df_pheno, kernel_type,
    centering_kernel=True,
    outer_subtract_absolute=True,
    double_centering=True,
    verbose=True,
    **kwargs
):
    """Kernel Transformation. Important preprocess to cross clustering. In this step, genes from different phenotypes are regarded as different genes. and the distance between them is computed. this function is decorated by :func:`MATTE.utils.kw_decorator`.

    :param df_exp: Expression dataframe.
    :type df_exp: pandas.DataFrame
    :param df_pheno: Phenotype dataframe.
    :type df_pheno: pandas.Series
    :param kernel_type: Kernel type, 'rbf' or 'localrbf' or 'mean'. functions are also allowed.
    :type kernel_type: str or function
    :param centering_kernel: whether to centering kernel, defaults to True
    :type centering_kernel: bool, optional
    :param outer_subtract_absolute: in outer subtract, use absolute or not, defaults to True
    :type outer_subtract_absolute: bool, optional
    :param double_centering: whether double centering kernel matrix or not, defaults to True
    :type double_centering: bool, optional
    :param verbose: defaults to True
    :type verbose: bool, optional
    :return: kernel matrix and mixed genes
    :rtype: dict
    """
    if verbose:
        printv(f"Calculating the kernel matrix using {kernel_type}",verbose=verbose)
        # if kernel_type not in ['mean','meanrbf']:
        #     printv(f'kernel type: {kernel_type} are not recommended for long running time and low score',verbose=verbose)

    genes = df_exp.index
    samples = df_exp.columns
    phenoes = df_pheno.unique()
    genes_mixed = pd.Series(
        [f"{i}@{j}" for j in phenoes for i in genes]
    )

    if kernel_type == 'meanrbf':
        gamma = kwargs.get('gamma', 2 / len(samples))
        if type(gamma) == list:
            assert len(gamma) == len(phenoes), "gamma should be a list of length equal to pheno if it is a list"
        printv(f'Setting gamma to {gamma}',verbose=verbose)
        Mat = meanrbf_Kernel_transform(
            df_exp, df_pheno, genes_mixed, gamma, outer_subtract_absolute)

    elif kernel_type == 'mean':
        Mat = mean_Kernel_transform(
            df_exp, df_pheno, genes_mixed, outer_subtract_absolute, )
    
    elif kernel_type == "localrbf":
        K = kwargs.get('K',7)
        Mat =  localrbf_Kernel_transform(
            df_exp, df_pheno, genes_mixed,K, outer_subtract_absolute)

    elif type(kernel_type) == function:
        df_exp_mixed = generate_df_exp_mixed(
            df_exp, df_pheno).loc[genes_mixed, :]
        Mat = cdist_generic(df_exp_mixed.values, None, kernel_type)
    
    else:
        raise ValueError("The kernel type is not in the list.")
    
    if double_centering:
        Mat = Double_Centering(Mat)
    if centering_kernel:
        Mat = KernelCenterer().fit_transform(Mat)
    
    return Mat, genes_mixed




from .utils import printv, kw_decorator
from .preprocess import *
from sklearn.decomposition import PCA
import dill as pickle
import os
import pandas as pd
import numpy as np
import warnings
from functools import wraps
from .cluster import CrossCluster,build_results
from itertools import combinations
from tqdm import tqdm
warnings.filterwarnings('ignore')

__version__ = "1.0.3"
__all__ = ["PipeFunc", "AlignPipe", 'preprocess_funcs_generate','ModuleEmbedder']

class PipeFunc():
    """`PipeFunc` is a wrapper of a function with storing arguments and kwargs **but not run it** until calling it self. And is used in :class:`AlignPipe` as a step.
    """
    def __init__(self, func, name=None, *args, **kwargs) -> None:
        """Pipe function used in :class:`AlignPipe`

        :param func: base function
        :type func: function
        :param name: names show in `str` func, defaults to None
        :type name: str, optional

        """

        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.__name__ = func.__name__ if name is None else name
        self.__doc__ = func.__doc__

    def __str_generate(self) -> str:
        argtext = " ".join([self.__argstext(i) for i in self.args])
        kwargtext = ",".join(
            [f"{k}={self.__argstext(v)}" for k, v in self.kwargs.items()])
        return f"""<PipeFunc> {self.func.__name__}({argtext}{kwargtext})"""

    def __str__(self) -> str:
        return self._strtext if hasattr(self,"_strtext") else self.__str_generate()

    def __update_str(func):
        """update the :code:str(self)

        :param func: _description_
        :type func: _type_
        """
        @wraps(func) 
        def wrapper(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            self._strtext = self.__str_generate()
            return result
        return wrapper

    def __call__(self, *args, **kwargs):
        args = self.args + args
        kwargs = {**self.kwargs, **kwargs}
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return self.__str__()

    @__update_str
    def add_params(self, *args, **kwargs):
        """Add params to PipeFunc, and refresh `str` text
        """        
        self.args += args
        self.kwargs.update(kwargs)

    def __argstext(self, arg):
        if type(arg) in [pd.DataFrame, pd.Series, np.ndarray]:
            return f"{type(arg)}({arg.shape})"
        elif type(arg) == str:
            return '\''+str(arg) + '\''
        else:
            return str(arg)


class AlignPipe():
    """Basic class in MATTE. Stores a list of functions or transformers.
    One can see the pipeline in :func:`__str__`

    The funcstions in :attr:`funcs` are called in order, and the results are passed to the next function. All returns of the functions should be `dict`,(using :func:`MATTE.utils.kw_decorator` are recommended). Then :attr:`cluster_func` will be called to cluster the results.

    .. note:: 
        Use :func:`MATTE.AlignPipe.add_step` to add a function to :attr:`funcs`, but can change order or delete some functions like `list`.
    
    """    
    def __init__(self, init=True, **config) -> None:
        """Initialize a AlignPipe object.

        :param init: weather to set up to default pipeline or not, defaults to True
        :type init: bool, optional
        """        
        for k, v in config.items():
            setattr(self, k, v)
        self.funcs = []
        self.cluster_func = []
        if init:
            self.default_pipeline()
    def __str__(self) -> str:
        """str is a string that shows the pipeline.

        :return: _summary_str (containing functions and parameters)
        :rtype: str
        """        
        strtext = "MATTE calculation pipeline\n"
        for n, f in enumerate(self.funcs):
            strtext += f"## STEP {n} \t"
            strtext += str(f)+"\n"
        for n, f in enumerate(self.cluster_func):
            strtext += f"## CLUSTER STEP {n} \t"
            strtext += str(f)+"\n"
        return strtext

    def __repr__(self) -> str:
        return self.__str__()

    def add_param(self, **kwargs):
        """add a function that return parameters to :attr:`funcs`.
        Multiple parameters can be added at the same time.
        """        
        for name, parm in kwargs.items():
            @kw_decorator(kw=name)
            def add_param():
                f"""adding params {name}."""
                return parm

            new_func = PipeFunc(func=add_param, name=f"add_{name}")
            self.funcs.append(new_func)

    def add_step(self, func, *setting, **kwsetting):
        """Add a function to :attr:`funcs`

        :param func: function(should be decarated by :func:`kw_decorator`) to add
        :type func: function
        """        
        new_func = PipeFunc(func, *setting, **kwsetting)
        self.funcs.append(new_func)

    def add_transformer(self, transformer):
        """Add a transformer to :attr:`funcs`

        :param transformer: transformer to add
        :type transformer: objects that has `fit` and `transform` methods
        :raises TypeError: if transformer is not a transformer
        """        
        if (not hasattr(transformer, "fit")) or (not hasattr(transformer, "transform")):

            raise TypeError(
                "The object is not a transformer (containing fit and transform function).")

        self.funcs.append(transformer)

    def set_cluster_method(self, func, *setting, **kwsetting):
        """Add a function to :attr:`cluster_func`

        :param func: function(should be decarated by :func:`kw_decorator`) to add
        :type func: function
        """        
        self.cluster_func.append(PipeFunc(func=func, *setting, **kwsetting))

    def __generate_transform_function(self, calling, kw="before_cluster_df"):
        if calling == "fit":

            @kw_decorator(kw=kw)
            def fit(transformer, before_cluster_df):
                return transformer.fit_transform(before_cluster_df)

            return fit

        else:

            @kw_decorator(kw=kw)
            def transform(transformer, before_cluster_df):
                return transformer.transform(before_cluster_df)

            return transform

    def sub_pipe(self, funcs_index, cluster_methods_index):

        new_pipe = AlignPipe(init=False)
        new_pipe.funcs = self.funcs[funcs_index]
        new_pipe.cluster_func = self.cluster_func[cluster_methods_index]
        return new_pipe

    def transform(self, df_exp, df_pheno, saving_temp=False, verbose=True):
        """Transform the dataframe using the pipeline.
        All parameters are same to :func:`AlignPipe.fit_transform` except use `transform` instead of `calculate`.
        """        
        tmpt_result = {"df_exp": df_exp, "df_pheno": df_pheno}
        for f in self.funcs:
            if type(f) == PipeFunc:
                printv(f"Running function {f.__name__}", verbose=verbose)
                tmpt_result.update(f(**tmpt_result))
            else:
                printv(f"Tranforming using model {f}", verbose=verbose)
                _f = self.__generate_transform_function(calling="transform")
                tmpt_result.update(_f(transformer=f, **tmpt_result))

        if saving_temp:
            if not os.path.exists(saving_temp):
                os.mkdir(saving_temp)

            with open(os.path.join(saving_temp, 'tempt.result'), "wb") as tempf:
                pickle.dump(tmpt_result, tempf)

    def fit_transform(self, df_exp, df_pheno, saving_temp=False, verbose=True):
        """Fit the data using the pipeline. If first performed, use :func:`AlignPipe.calculate` instead.

        :param df_exp: expression data whose index are genes and columns are samples
        :type df_exp: `pandas.DataFrame`
        :param df_pheno: phenotype data whose index are samples
        :type df_pheno: `pandas.Series`
        :param saving_temp: whether to save temp or not.if not False,set the parameter to be saving dir, defaults to False
        :type saving_temp: bool or str, optional
        :param verbose: defaults to True
        :type verbose: bool, optional
        :return: final results containing all results of the pipeline functions
        :rtype: dict
        """        

        tmpt_result = {"df_exp": df_exp,
                       "df_pheno": df_pheno, "verbose": verbose}
        for f in self.funcs:
            if type(f) == PipeFunc:
                printv(f"Running function {f.__name__}", verbose=verbose)
                f_res = f(**tmpt_result)
                tmpt_result.update(f_res)
            else:
                printv(f"Tranforming using model {f}", verbose=verbose)
                _f = self.__generate_transform_function(calling="fit")
                tmpt_result.update(_f(transformer=f, **tmpt_result))

        if saving_temp:
            printv(f"Saving tmpt result in {saving_temp}", verbose=verbose)
            if not os.path.exists(saving_temp):
                os.mkdir(saving_temp)
            with open(os.path.join(saving_temp, 'tempt.result'), "wb") as tempf:
                pickle.dump(tmpt_result, tempf)

        return tmpt_result

    def calculate(self, df_exp, df_pheno, verbose=True):
        """Calculate the data using the pipeline.

        :param df_exp: expression data whose index are genes and columns are samples
        :type df_exp: `pandas.DataFrame`
        :param df_pheno: phenotype data whose index are samples
        :type df_pheno: `pandas.Series`
        :param verbose: defaults to True
        :type verbose: bool, optional
        :return: clustering results
        :rtype: :class:`MATTE.analysis.ClusterResult`
        """        

        tmpt_result = self.fit_transform(
            df_exp, df_pheno, verbose=verbose, saving_temp=False)

        return self.calculate_from_temp(tmpt_result, verbose)

    def calculate_from_temp(self, tmpt_result, verbose=True):
        """Calculate the data using the pipeline but from temp file.

        :param tmpt_result: temp result saved by :func:`MATTE.AlignPipe.fit_transform`
        :type tmpt_result: dict
        :param verbose: defaults to True
        :type verbose: bool, optional
        :return: clustering results
        :rtype: :class:`MATTE.analysis.ClusterResult`
        """        
        for f in self.cluster_func:
            printv(f"Running function {f.__name__}", verbose=verbose)
            f_res = f(**tmpt_result)
            tmpt_result.update(f_res)

        result = tmpt_result["Result"]
        result.cluster_properties.update({'Process': self.__str__()})
        return result

    def get_attribute_from_transformer(self, attribute):
        """Get the attribute from the transformer.

        :param attribute: attribute name
        :type attribute: str
        :return: attribute value
        :rtype: any
        """        
        return next(
            (
                getattr(i, attribute)
                for i in self.funcs
                if type(i) != PipeFunc and hasattr(i, attribute)
            ),
            None,
        )

    def default_pipeline(self):
        """Get the default pipeline.

        :return: default pipeline
        :rtype: :class:`AlignPipe`

        .. note:: default pipeline contrains **preprocessing**, and if data is preprocessed, delete it.
        
        see default pipeline, run code below to see the result. 

        .. code-block:: python

            import MATTE
            print(MATTE.AlignPipe())

        """        

        self.add_step(inputs_check)
        self.add_step(RPKM2TPM)
        self.add_step(log2transform)
        self.add_step(exp_filter, gene_filter=None)
        self.add_step(
            Kernel_transform,
            kernel_type="mean", centering_kernel=True,
            outer_subtract_absolute=True, double_centering=True
        )
        self.add_transformer(
            PCA(n_components=16))

        @kw_decorator(kw='weights')
        def adding_weights():
            return self.get_attribute_from_transformer('explained_variance_')

        self.add_step(
            func=adding_weights)
        clustering_func = CrossCluster(preset='kmeans',n_clusters=8, method="a", dist_type="a",n_iters=20)
        self.set_cluster_method(
            func=clustering_func,
        )
        self.set_cluster_method(func=build_results)

        return self



class ModuleEmbedder():
    """MATTE embedder. This class is used to embeding the data using MATTE pipeline.

    There are two types of embedding:
    
    1. feature selection: ranking gene by its importance sum in each pair of phenotypes comparasion.
    2. feature extraction: using pca to transfrom data to most variant space 

    .. note:: 
        Inputs of ModuleEmbedder is not the same as pipeline. Row of Expression data is sample, column is gene. 
    """    
    def __init__(self, pipeline=None) -> None:
        """Initialize the MATTE embedder.

        :param pipeline: pipeline used to each pair of phenotypes, defaults to None(default pipeline)
        :type pipeline: :class:`MATTE.AlignPipe`, optional

        .. note:: default pipeline contrains **preprocessing**, and if data is preprocessed, delete it.
        """        
        self.pipeline = AlignPipe() if pipeline is None else pipeline
        self.cluster_res = []

    def _pipeline_clustering_single(self, X, y,):
        CR = self.pipeline.calculate(df_exp=X.T, df_pheno=y, verbose=False)
        self.cluster_res.append(CR)
        return CR

    def pipeline_clustering(self, X, y, verbose=True):
        """Clustering the data using pipeline for each pair of phenotypes.
        A progress bar is set to inform the user about the progress.

        :param X: expression data whose index are samples and columns are genes
        :type X: `pandas.DataFrame`
        :param y: phenotype data whose index are samples
        :type y: `pandas.Series`
        :param verbose: defaults to True
        :type verbose: bool, optional
        """        
        bar = tqdm(total=len(list(combinations(self.labels, 2))))
        for idx, (i, j) in enumerate(combinations(self.labels, 2)):
            bar.set_description(f'round {idx}: {i} vs {j}')
            X_ij = X[(y == j) | (y == i)]
            y_ij = y[(y == i) | (y == j)]
            CR = self._pipeline_clustering_single(X_ij, y_ij,)
            self.cluster_res.append(CR)
            bar.update(1)
        bar.close()

    def _gene_rank_single(self, CR, X):
        """Rank genes by their importance sum in one pair of phenotypes comparasion.
        Importance is calculated by the module SNR, see :func:`MATTE.analysis.ClusterResult.ModuleSNR`.

        :param CR: clustering results
        :type CR: :class:`MATTE.analysis.ClusterResult`
        :param X: expression data whose index are samples and columns are genes
        :type X: `pandas.DataFrame`
        :return: ranked genes
        :rtype: `pandas.Series`
        """        
        sf = CR.SampleFeature()
        snr = CR.ModuleSNR(sf)
        importance = pd.Series(data=0.0, index=X.columns)
        for module,genes in CR.module_genes.items():
            importance[genes] = snr[f'{module}_0']
        return importance

    def _transform_single(self, CR, thres=0.85):
        """transform in single pair of phenotypes comparasion.

        :param CR: clustering results
        :type CR: :class:`MATTE.analysis.ClusterResult`
        :param thres: thereshold used in selecting genes to :class:`sklearn.decomposition.PCA`,(sum of select genes' explain variant ratio reach `thres`) defaults to 0.85
        :type thres: float, optional
        :return: transformer 
        :rtype: :class:`sklearn.decomposition.PCA`
        """        
        sf = CR.SampleFeature()
        snr = CR.ModuleSNR(sf)
        # if not (snr >= 1.5).any():
        #     return None
        i = 0
        while True:
            module = snr.index[i].split('_')[0]
            genes = CR.module_genes[module]
            if len(genes) > 1:
                break
            i += 1

        pca = PCA(n_components=None)
        X = CR.df_exp.loc[genes, :].T
        pca.fit(X)
        ## sort genes by explained variance; 
        # then select the top genes until the explained variance is larger than thres
        exp_vr = pca.explained_variance_ratio_
        select_genes_idx =  np.argsort(exp_vr)[::-1][:np.where(np.cumsum(np.sort(exp_vr)[::-1])>thres)[0][0]]
        select_genes = pca.feature_names_in_[select_genes_idx]

        pca = PCA(n_components=None)
        X = CR.df_exp.loc[select_genes, :].T
        pca.fit(X)
        return pca

    def transform(self, X):
        """transform the data using pipeline for each pair of phenotypes.
        Run :func:`MATTEEmbedder.transform_fit` first.

        :param X: expression data whose index are samples and columns are genes
        :type X: `pandas.DataFrame`
        :return: transformed data
        :rtype: `pandas.DataFrame`
        """        
        assert hasattr(self, 'transformers'), 'Run transform_fit first.'
        Xs = []
        used_genes = set()
        for transformer in self.transformers:
            features = transformer.feature_names_in_
            used_genes = set(features) | used_genes
            Xs.append(transformer.transform(X.loc[:, features]))
        Xs = np.concatenate(Xs, axis=1)
        self.used_genes = list(used_genes)
        return Xs

    def gene_rank(self, X: pd.DataFrame, y, verbose=True):
        """ranking genes by their :func:`MATTE.analysis.ClusterResult.ModuleSNR` in each phenotype pairs.
        Each gene in one phenotype is sum to 1000. Meanwhile, set :attr:`ModuleEmbedder.gene_ranking_sep` to calculte each phenotype score.

        :param X: Expression data whose index are samples and columns are genes.
        :type X: `pd.DataFrame`
        :param y: phenotypes whose index are samples.
        :type y: `pd.Series`
        :param verbose: defaults to True
        :type verbose: bool, optional
        :return: gene ranking score
        :rtype: `pd.Series`
        """
        self.labels = np.unique(y)
        if len(self.cluster_res) == 0:
            self.pipeline_clustering(X, y, verbose=verbose)
        if verbose:
            print(f"There are {len(self.labels)} labels: {self.labels}")

        importance_template = pd.Series(data=0.0, index=X.columns)
        clusters_importance = {}
        for idx, (i, j) in enumerate(combinations(self.labels, 2)):
            CR = self.cluster_res[idx]
            X_ij = pd.concat([X[y == i], X[y == j]])
            importance = self._gene_rank_single(CR, X_ij)
            clusters_importance[i] = clusters_importance.get(i,importance_template.copy()) + importance
            clusters_importance[j] = clusters_importance.get(j,importance_template.copy()) + importance
        # overall importance is the sum of the cluster importance
        gene_ranking = pd.Series(data=0.0, index=X.columns)
        for cluster, importance in clusters_importance.items():
            gene_ranking += 1000*importance/importance.sum()
        gene_ranking /= len(self.labels)
        self.gene_ranking = gene_ranking
        self.gene_ranking_sep = clusters_importance
        return gene_ranking

    def transform_fit(self, X: pd.DataFrame, y, verbose=True):
        """transform the data using pipeline for each pair of phenotypes.

        :param X: expression data whose index are samples and columns are genes
        :type X: `pandas.DataFrame`
        :param y: phenotypes whose index are samples.
        :type y: `pandas.Series`
        :param verbose: defaults to True
        :type verbose: bool, optional
        :return: self
        :rtype: :class:`ModuleEmbedder`
        """        
        self.transformers = []
        self.labels = np.unique(y)
        if verbose:
            print(f"There are {len(self.labels)} labels: {self.labels}")

        if len(self.cluster_res) == 0:
            self.pipeline_clustering(X, y, verbose=verbose)

        for idx, (i, j) in enumerate(combinations(self.labels, 2)):
            CR = self.cluster_res[idx]
            transformer = self._transform_single(CR)
            if transformer is not None:
                self.transformers.append(transformer)

        return self

    def save(self, save_path):
        """saving the embedder using `dill`

        :param save_path: path to save the embedder
        :type save_path: str
        """        
        with open(save_path, 'wb') as f:
            pickle.dump(self, f)

def make_pipefunc(func,kw='df_exp'):

    @kw_decorator(kw=kw)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return PipeFunc(wrapper, name=func.__name__)

def package_test(n_genes=100, n_samples=100, pipe=None,seed=0,verbose=True):
    """testing the package by generating data.

    :param n_genes: number of genes, defaults to 100
    :type n_genes: int, optional
    :param n_samples: number of samples, defaults to 100
    :type n_samples: int, optional
    :param pipe: pipeline to use, defaults to None(default pipeline)
    :type pipe: :class:`MATTE.AlignPipe`, optional
    :param seed: random seed, defaults to 0
    :type seed: int, optional
    :param verbose: defaults to True, print all the information
    :type verbose: bool, optional
    :return: clustering result
    :rtype: :class:`MATTE.analysis.ClusterResult`
    """    
    np.random.seed(seed)
    ## generate a test data
    genes = [f"gene{i}" for i in range(n_genes)]
    samples = [f"sample{i}" for i in range(n_samples)]
    base_exp = np.random.gamma(3,1,(n_genes))
    ## from base_exp select 20% to change to a new value
    disturb_exp = base_exp.copy()
    disturb_exp[0:int(n_genes/5)] = np.random.gamma(3,1,int(n_genes/5))
    
    df_pheno = pd.Series(
        [f"P{i}" for i in np.random.randint(size=n_samples, low=0, high=2)],
        index=samples,)
    df_exps = []
    for i in range(2):
        sub_samples = df_pheno[df_pheno == f"P{i}"].index
        df_exp = pd.DataFrame(
            np.random.normal(
                loc=[base_exp,disturb_exp][i],
                size=(sub_samples.size,n_genes)),
            index=sub_samples,
            columns=genes,
        ).T
        df_exps.append(np.abs(df_exp))
    df_exp = pd.concat(df_exps, axis=1)

    ## generate a test pipe
    pipe = AlignPipe(init=True) if pipe is None else pipe
    R = pipe.calculate(df_exp, df_pheno, verbose=verbose)
    if verbose:
        print(R.summary())
    return R,{'df_exp':df_exp,'df_pheno':df_pheno}



from setuptools import setup,find_packages

## Adding following code if Requirements install needed.
# import os
# os.system("pip3 install -r requirements.txt")

requires_packages = [
    "sklearn", "numpy", "biopython", "tqdm", "pandas", "scipy", "statsmodels",
    "matplotlib","seaborn","joblib","dill"]

setup(
    name = 'MATTE',
    version = '1.0.3',
    packages = find_packages(),
    install_requires = requires_packages,
    url = 'https://github.com/gx-Cai/MATTE',
    author = 'Cai Guoxin',
    author_email = 'gxcai@zju.edu.cn',
    include_package_data=True
)
